package com.ibm.eis.mdrr.query.server;

import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import org.w3c.dom.Element;
import org.w3c.dom.traversal.NodeIterator;

/**
 * A template evaluator that may be used by the XSLT query transform.
 * It may be subclassed to provide application-specific evaluation logic.
 */
public class DefaultTemplateEvaluator implements ITemplateEvaluator {
	
	protected ITemplateEvaluatorFactory _templateFactory = null;

	protected Map<String, IItemTemplateEvaluator> _itemTemplates = new HashMap<String, IItemTemplateEvaluator>();
	protected Map<String, IRelationshipTemplateEvaluator> _relationshipTemplates = new HashMap<String, IRelationshipTemplateEvaluator>();
	protected Map<String, Set<IItem>> _itemResults = null;
	protected Map<String, Set<IRelationship>> _relationshipResults = null;
	
	/**
	 * @param templateFactory an application-specific factory class that can create item/relationship template evaluators
	 */
	public DefaultTemplateEvaluator(ITemplateEvaluatorFactory templateFactory) {
		_templateFactory = templateFactory;
	}

	/**
	 * This is a simple template evaluation logic that rougly duplicates the
	 * effect of the current framework. 
	 */
	protected void evaluateAll() throws CMDBfException {
		_itemResults = new HashMap<String, Set<IItem>>();
		_relationshipResults = new HashMap<String, Set<IRelationship>>();

		// calculate all item template results, disrerding relationship requirements
		for (String itemTemplateName : _itemTemplates.keySet()) {
			_itemResults.put(itemTemplateName, _itemTemplates.get(itemTemplateName).evaluate());
		}

		// calculate relationship results, pruning item template results after each relationship calculation
		for (String relationshipTemplateName : _relationshipTemplates.keySet()) {
			IRelationshipTemplateEvaluator relationshipTemplate = _relationshipTemplates.get(relationshipTemplateName);
			String sourceTemplateName = relationshipTemplate.getSourceTemplateName();
			Set<IItem> sourceResults = _itemResults.get(sourceTemplateName);
			String targetTemplateName = relationshipTemplate.getTargetTemplateName();
			Set<IItem> targetResults = _itemResults.get(targetTemplateName);
			Set<IRelationship> relationshipResults = _relationshipTemplates.get(relationshipTemplateName).evaluate(sourceResults, targetResults);

			// update item template results, retaining only items that are found in the relationship result
			Set<IItem> foundSources = new HashSet<IItem>();
			Set<IItem> foundTargets = new HashSet<IItem>();
			for (IRelationship relationship : relationshipResults) {
				foundSources.add(relationship.getSource());
				foundTargets.add(relationship.getTarget());
			}
			sourceResults.retainAll(foundSources);
			targetResults.retainAll(foundTargets);
			
			_relationshipResults.put(relationshipTemplateName, relationshipResults);
		}
		
		// remove all the relationships where the source or target is no longer in the item template result
		for (String relationshipTemplateName : _relationshipTemplates.keySet()) {
			IRelationshipTemplateEvaluator relationshipTemplate = _relationshipTemplates.get(relationshipTemplateName);
			String sourceTemplateName = relationshipTemplate.getSourceTemplateName();
			Set<IItem> sourceResults = _itemResults.get(sourceTemplateName);
			String targetTemplateName = relationshipTemplate.getTargetTemplateName();
			Set<IItem> targetResults = _itemResults.get(targetTemplateName);
			Set<IRelationship> relationshipResults = _relationshipResults.get(relationshipTemplateName);
			for (IRelationship relationship : relationshipResults) {
				if (   !sourceResults.contains(relationship.getSource())
					|| !targetResults.contains(relationship.getTarget()))
				{
					relationshipResults.remove(relationship);
				}
			}			
		}
	}
	
	public IItemTemplateEvaluator createItemTemplateEvaluator(NodeIterator itemTemplateIterator) throws CMDBfException {
		Element itemTemplate = (Element) itemTemplateIterator.nextNode();
		String templateName = itemTemplate.getAttribute("id");
		IItemTemplateEvaluator itemQuery = _templateFactory.createItemTemplateEvaluator(itemTemplate);
		_itemTemplates.put(templateName, itemQuery);
		return itemQuery;
	}
	
	public List<IItem> getItemTemplateResults(String templateName) throws CMDBfException {
		if (_itemResults == null)
			evaluateAll();
		return Arrays.asList(_itemResults.get(templateName).toArray(new IItem[]{}));
	}
	
	public IRelationshipTemplateEvaluator createRelationshipTemplateEvaluator(NodeIterator relationshipTemplateIterator) throws CMDBfException {
		Element relationshipTemplate = (Element) relationshipTemplateIterator.nextNode();
		String templateName = relationshipTemplate.getAttribute("id");
		IRelationshipTemplateEvaluator relationshipQuery = _templateFactory.createRelationshipTemplateEvaluator(relationshipTemplate);
		_relationshipTemplates.put(templateName, relationshipQuery);
		return relationshipQuery;
	}
	
	public List<IRelationship> getRelationshipTemplateResults(String templateName) throws CMDBfException {
		if (_relationshipResults == null)
			evaluateAll();
		return Arrays.asList(_relationshipResults.get(templateName).toArray(new IRelationship[]{}));
	}
}
