<?xml version="1.0" encoding="UTF-8"?>

<xsl:stylesheet xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
    version="1.0"
    xmlns:cmdbf="http://cmdbf.org/schema/1-0-0/datamodel"
    xmlns:xslt="http://xml.apache.org/xslt"
    xmlns:xalan="http://xml.apache.org/xalan"
	xmlns:AdditionalRecordType="xalan://com.ibm.eis.mdrr.query.server.AdditionalRecordType"
	xmlns:IItem="xalan://com.ibm.eis.mdrr.query.server.IItem"
	xmlns:InstanceId="xalan://com.ibm.eis.mdrr.query.server.InstanceId"
	xmlns:IRecord="xalan://com.ibm.eis.mdrr.query.server.IRecord"
	xmlns:IRelationship="xalan://com.ibm.eis.mdrr.query.server.IRelationship"
	xmlns:ITemplateEvaluator="xalan://com.ibm.eis.mdrr.query.server.ITemplateEvaluator"
	xmlns:ListUtil="xalan://com.ibm.eis.mdrr.query.server.ListUtil"
    exclude-result-prefixes="xslt xalan AdditionalRecordType IItem InstanceId IRecord IRelationship ITemplateEvaluator ListUtil">

    <xsl:output method="xml" indent="yes" xalan:indent-amount="3"/>
    
    <!-- This parameter must contain an instance of an ITemplateEvalutor -->
    <xsl:param name="templateEvaluator"/>
	
    <!-- 
    	Template processes a <cmdbf:query> by first creating a template evaluator for each item/relationship template,
    	then evaluating and outputting the results of each template.
     -->
    <xsl:template match="cmdbf:query">
	   	<xsl:apply-templates select="*" mode="createQuery"/>
    	<cmdbf:queryResponse>
	    	<xsl:apply-templates select="cmdbf:itemTemplate" mode="evalQuery"/>
	    	<xsl:apply-templates select="cmdbf:relationshipTemplate" mode="evalQuery"/>
    	</cmdbf:queryResponse>
    </xsl:template>
    
    <!-- 
    	Template to create an application-specific itemTemplateEvaluator from a <cmdbf:itemTemplate> element
     -->
    <xsl:template match="cmdbf:itemTemplate" mode="createQuery">
    	<xsl:variable name="query" select="ITemplateEvaluator:createItemTemplateEvaluator($templateEvaluator, .)"/>
    </xsl:template>

    <!-- 
    	Template to create an application-specific relationshipTemplateEvaluator from a <cmdbf:relationshipTemplate> element
     -->
    <xsl:template match="cmdbf:relationshipTemplate" mode="createQuery">
    	<xsl:variable name="query" select="ITemplateEvaluator:createRelationshipTemplateEvaluator($templateEvaluator, .)"/>
	</xsl:template>
    
    <!-- 
    	Template to output the results of each itemTemplate evaluation as a <cmdbf:ndoes> element
    	The list of items associated with each item template is output by the nodes template.
     -->
    <xsl:template match="cmdbf:itemTemplate" mode="evalQuery">
		<xsl:if test="not(@suppressFromResult='true') and not(@suppressFromResult='1')">
	    	<xsl:variable name="itemList" select="ITemplateEvaluator:getItemTemplateResults($templateEvaluator, @id)"/>
    		<xsl:if test="not(ListUtil:isEmpty($itemList))">
	    		<cmdbf:nodes templateId="{@id}">
	    			<xsl:call-template name="nodes">
	    				<xsl:with-param name="itemList" select="$itemList"/>
		    		</xsl:call-template>
		    	</cmdbf:nodes>
		   	</xsl:if>
	    </xsl:if>
    </xsl:template>

    <!-- 
    	Template to output a list of items from an itemTemplate evaluation as <cmdbf:item> elements
    	The list of records associated with each item is output by the itemRecords template;
    	the list of instance ids is output by the instanceIds template;
    	and the list of additional record types is output by the additionalRecordTypes template.
     -->
    <xsl:template name="nodes">
		<xsl:param name="itemList"/>
    	<xsl:if test="not(ListUtil:isEmpty($itemList))">
    		<xsl:variable name="item" select="ListUtil:head($itemList)"/>
	    	<cmdbf:item>
   				<xsl:call-template name="itemRecords">
   					<xsl:with-param name="itemRecordList" select="IItem:getRecords($item)"/>
   				</xsl:call-template>
   				<xsl:call-template name="instanceIds">
   					<xsl:with-param name="instanceIdList" select="IItem:getInstanceIds($item)"/>
   				</xsl:call-template>
   				<xsl:call-template name="additionalRecordTypes">
   					<xsl:with-param name="additionalRecordTypeList" select="IItem:getAdditionalRecordTypes($item)"/>
   				</xsl:call-template>
    		</cmdbf:item>
    		<xsl:call-template name="nodes">
    			<xsl:with-param name="itemList" select="ListUtil:tail($itemList)"/>
    		</xsl:call-template>
    	</xsl:if>
    </xsl:template>
    
    <!-- 
    	Template to output each record associated with an item as a <cmdbf:record> element
    	(This is implemented separately from the relationshipRecords template below because these application generally format
    		these records quite differently.)
    	Each record is passed to an application-specific itemRecord template, which must be supplied by the framework user.
		Each record is also passed to an itemRecordMetadata template, whose default empty implementation may be overridden.
     -->
    <xsl:template name="itemRecords">
		<xsl:param name="itemRecordList"/>
		<xsl:if test="not(ListUtil:isEmpty($itemRecordList))">
			<xsl:variable name="itemRecord" select="ListUtil:head($itemRecordList)"/>
			<cmdbf:record>
				<xsl:call-template name="itemRecord">
					<xsl:with-param name="itemRecord" select="$itemRecord"/>
				</xsl:call-template>
				<cmdbf:recordMetadata>
					<cmdbf:recordId><xsl:value-of select="IRecord:getRecordId($itemRecord)"/></cmdbf:recordId>
					<xsl:call-template name="itemRecordMetadata">
						<xsl:with-param name="itemRecord" select="$itemRecord"/>
					</xsl:call-template>
				</cmdbf:recordMetadata>
			</cmdbf:record>
			<xsl:call-template name="itemRecords">
				<xsl:with-param name="itemRecordList" select="ListUtil:tail($itemRecordList)"/>
			</xsl:call-template>
		</xsl:if>    
    </xsl:template>

    <!-- 
    	Template to output each element of a non-empty list of instance ids as a <cmdbf:instanceId> element.
		It is called from the nodes and edges templates.
     -->
    <xsl:template name="instanceIds">
    	<xsl:param name="instanceIdList"/>
		<cmdbf:instanceId>
	    	<xsl:call-template name="instanceId">
    			<xsl:with-param name="instanceId" select="ListUtil:head($instanceIdList)"/>
	    	</xsl:call-template>
	    </cmdbf:instanceId>
		<xsl:variable name="instanceIdListTail" select="ListUtil:tail($instanceIdList)"/>
		<xsl:if test="not(ListUtil:isEmpty($instanceIdListTail))">
			<xsl:call-template name="instanceIds">
				<xsl:with-param name="instanceIdList" select="$instanceIdListTail"/>
			</xsl:call-template>
		</xsl:if>    	
    </xsl:template>
    
    <!-- 
    	Template to output a single instance id as a sequence of <cmdbf:mdrId> and <cmdbf:localId> elements
		It is called from the nodes, edges, and instanceIds templates.
     -->
    <xsl:template name="instanceId">
    	<xsl:param name="instanceId"/>
		<cmdbf:mdrId><xsl:value-of select="InstanceId:getMdrId($instanceId)"/></cmdbf:mdrId>
		<cmdbf:localId><xsl:value-of select="InstanceId:getLocalId($instanceId)"/></cmdbf:localId>
    </xsl:template>

    <!-- 
    	Template to output each of a (possibly null or empty) list of additionalRecordTypes as <cmdbf:additionalRecordType> elements
		It is called from the nodes and edges templates.
     -->
    <xsl:template name="additionalRecordTypes">
    	<xsl:param name="additionalRecordTypeList"/>
    	<xsl:if test="not(ListUtil:isEmpty($additionalRecordTypeList))">
    		<xsl:variable name="additionalRecordType" select="ListUtil:head($additionalRecordTypeList)"/>
    		<cmdbf:additionalRecordType namespace="{AdditionalRecordType:getNamespace($additionalRecordType)}"
    				localName="{AdditionalRecordType:getLocalName($additionalRecordType)}"/>
			<xsl:call-template name="additionalRecordTypes">
				<xsl:with-param name="additionalRecordTypeList" select="ListUtil:tail($additionalRecordTypeList)"/>
			</xsl:call-template>
    	</xsl:if>
    </xsl:template>
    
	<!--
		Default item record metadata format template
		May be overridden by application-specific implementation
	 -->
    <xsl:template name="itemRecordMetadata">
    	<xsl:param name="itemRecord"/>
    </xsl:template>
    
    <!-- 
    	Template to output the results of each relationshipTemplate evaluation as a <cmdbf:edges> element
    	The list of relationships associated with each relationship template is output by the edges template.
     -->
    <xsl:template match="cmdbf:relationshipTemplate" mode="evalQuery">
		<xsl:if test="not(@suppressFromResult='true') and not(@suppressFromResult='1')">
    		<xsl:variable name="relationshipList" select="ITemplateEvaluator:getRelationshipTemplateResults($templateEvaluator, @id)"/>
	    	<xsl:if test="not(ListUtil:isEmpty($relationshipList))">
		    	<cmdbf:edges templateId="{@id}">
	    			<xsl:call-template name="edges">
	    				<xsl:with-param name="relationshipList" select="$relationshipList"/>
		    		</xsl:call-template>
			    </cmdbf:edges>
	    	</xsl:if>
	    </xsl:if>
	</xsl:template>
    
    <!-- 
    	Template to output a list of relationships from a RelationshipTemplate evaluation as <cmdbf:relationship> elements
    	The list of records associated with each relationship is output by the relationshipRecords template.
     -->
	<xsl:template name="edges">
		<xsl:param name="relationshipList"/>
    	<xsl:if test="not(ListUtil:isEmpty($relationshipList))">
    		<xsl:variable name="relationship" select="ListUtil:head($relationshipList)"/>
	    	<cmdbf:relationship>
    			<cmdbf:source>
    				<xsl:call-template name="instanceId">
    					<xsl:with-param name="instanceId" select="ListUtil:head(IItem:getInstanceIds(IRelationship:getSource($relationship)))"/>
    				</xsl:call-template>
    			</cmdbf:source>
    			<cmdbf:target>
    				<xsl:call-template name="instanceId">
    					<xsl:with-param name="instanceId" select="ListUtil:head(IItem:getInstanceIds(IRelationship:getTarget($relationship)))"/>
    				</xsl:call-template>
    			</cmdbf:target>
   				<xsl:call-template name="relationshipRecords">
   					<xsl:with-param name="relationshipRecordList" select="IRelationship:getRecords($relationship)"/>
   				</xsl:call-template>
   				<xsl:call-template name="instanceIds">
   					<xsl:with-param name="instanceIdList" select="IItem:getInstanceIds($relationship)"/>
   				</xsl:call-template>
   				<xsl:call-template name="additionalRecordTypes">
   					<xsl:with-param name="additionalRecordTypeList" select="IItem:getAdditionalRecordTypes($relationship)"/>
   				</xsl:call-template>
    		</cmdbf:relationship>
    		<xsl:call-template name="edges">
    			<xsl:with-param name="relationshipList" select="ListUtil:tail($relationshipList)"/>
    		</xsl:call-template>
    	</xsl:if>
	</xsl:template>
    
    <!-- 
    	Template to output each record associated with a relationship as a <cmdbf:record> element
    	Each record is passed to an application-specific relationshipRecord template, which must be supplied by the framework user.
		Each record is also passed to a relationshipRecordMetadata template, whose default empty implementation may be overridden.
     -->
    <xsl:template name="relationshipRecords">
		<xsl:param name="relationshipRecordList"/>
		<xsl:if test="not(ListUtil:isEmpty($relationshipRecordList))">
			<xsl:variable name="relationshipRecord" select="ListUtil:head($relationshipRecordList)"/>
			<cmdbf:record>
				<xsl:call-template name="relationshipRecord">
					<xsl:with-param name="relationshipRecord" select="$relationshipRecord"/>
				</xsl:call-template>
				<cmdbf:recordMetadata>
					<cmdbf:recordId><xsl:value-of select="IRecord:getRecordId($relationshipRecord)"/></cmdbf:recordId>
					<xsl:call-template name="relationshipRecordMetadata">
						<xsl:with-param name="relationshipRecord" select="$relationshipRecord"/>
					</xsl:call-template>
				</cmdbf:recordMetadata>
			</cmdbf:record>
    		<xsl:call-template name="relationshipRecords">
    			<xsl:with-param name="relationshipRecordList" select="ListUtil:tail($relationshipRecordList)"/>
    		</xsl:call-template>			
		</xsl:if>    
    </xsl:template>

	<!--
		Default relationship record metadata output template
		May be overridden by an application-specific implementation
	 -->
    <xsl:template name="relationshipRecordMetadata">
    	<xsl:param name="relationshipRecord"/>
    </xsl:template>
	
</xsl:stylesheet>